import string
from datetime import datetime
from io import StringIO

import time_machine
from django.contrib.auth import get_user
from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import SimpleTestCase, TestCase, override_settings
from django.urls import NoReverseMatch, reverse
from django.utils.http import int_to_base36
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.django import TestCase as HypothesisTestCase
from management_auth.tokens import ManagementAuthTokenGenerator

token_strategy = st.text(
    alphabet=st.characters(
        codec="ascii", exclude_characters={"?", "#", *string.whitespace}
    ),
    min_size=3,
)


class LoginAsViewTestCase(HypothesisTestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create(username="user-1")

        cls.generator = ManagementAuthTokenGenerator()

    def assertLoginSuccess(self, response):  # noqa:N802
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/accounts/profile/")
        self.assertTrue(get_user(self.client).is_authenticated)

    def assertLoginFail(self, response):  # noqa:N802
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/accounts/login/?next=")
        self.assertFalse(get_user(self.client).is_authenticated)

    def test_authenticates_user(self):
        token = self.generator.make_token(self.user.id)

        last_login = self.user.last_login

        with self.assertNumQueries(10):
            response = self.client.get(
                reverse("management_auth_login_as", args=[token])
            )

        self.user.refresh_from_db()

        self.assertLoginSuccess(response)
        self.assertEqual(get_user(self.client), self.user)
        self.assertEqual(self.user.last_login, last_login)

    def test_disallow_head_request(self):
        with self.assertNumQueries(0):
            response = self.client.head(
                reverse("management_auth_login_as", args=["not-a-token"])
            )

        self.assertEqual(response.status_code, 405)

    @given(token_strategy)
    def test_invalid_token(self, token):
        # Work around the reverse potentially failing
        url = reverse("management_auth_login_as", args=["token-placeholder"]).replace(
            "token-placeholder", token
        )

        with self.assertNumQueries(0):
            response = self.client.get(url)

        if response.status_code == 404:
            # If it was a 404, ensure it's because of URL resolving
            with self.assertRaises(NoReverseMatch):
                reverse("management_auth_login_as", args=[token])
        else:
            self.assertLoginFail(response)

    def test_authenticates_logged_in_user(self):
        user_2 = User.objects.create(username="user-2")
        self.client.force_login(user_2)
        self.assertEqual(get_user(self.client), user_2)

        token = self.generator.make_token(self.user.id)

        with self.assertNumQueries(10):
            response = self.client.get(
                reverse("management_auth_login_as", args=[token])
            )

        self.assertLoginSuccess(response)
        self.assertEqual(get_user(self.client), self.user)

    def test_no_cache(self):
        token = self.generator.make_token(self.user.id)
        response = self.client.get(reverse("management_auth_login_as", args=[token]))

        self.assertEqual(
            response["Cache-Control"],
            "max-age=0, no-cache, no-store, must-revalidate, private",
        )


@time_machine.travel(datetime(2023, 1, 1))
class TokenGeneratorTestCase(SimpleTestCase):
    def setUp(self):
        self.user = User(pk=1)
        self.user_2 = User(pk=2)

        self.generator = ManagementAuthTokenGenerator()

    def test_creates_token(self):
        self.assertEqual(
            self.generator.make_token(self.user.id),
            "1o-WzEsNjBd:1pBllg:6iGF6Rjv0I24e3Y0y-G_pXSaNg8UzExOHfeHglQ2UCo",
        )

    def test_validates_token(self):
        token = self.generator.make_token(self.user.id)

        self.assertEqual(self.generator.get_user_from_token(token), self.user.id)

    def test_incorrect_timestamp(self):
        token = self.generator.make_token(self.user.id)

        timestamp, token = token.split(self.generator.timeout_sep, 1)

        timestamp = int_to_base36(15)

        self.assertIsNone(
            self.generator.get_user_from_token(
                self.generator.timeout_sep.join([timestamp, token])
            )
        )

    def test_no_token(self):
        self.assertIsNone(self.generator.get_user_from_token(""))

    def test_no_timestamp(self):
        token = self.generator.make_token(self.user.id)

        _, token = token.split(self.generator.timeout_sep, 1)

        self.assertIsNone(self.generator.get_user_from_token(token))

    def test_expired_token(self):
        token = self.generator.make_token(self.user.id)

        with time_machine.travel(datetime(2023, 1, 2)):
            self.assertIsNone(self.generator.get_user_from_token(token))

    @given(token_strategy)
    def test_invalid_token(self, token):
        self.assertIsNone(self.generator.get_user_from_token(token))

    def test_stable_tokens(self):
        self.assertEqual(
            self.generator.make_token(self.user.id),
            self.generator.make_token(self.user.id),
        )

    def test_different_tokens_for_different_users(self):
        self.assertNotEqual(
            self.generator.make_token(self.user.id),
            self.generator.make_token(self.user_2.id),
        )

    def test_different_token_for_different_timeout(self):
        self.assertNotEqual(
            self.generator.make_token(self.user.id),
            self.generator.make_token(self.user.id, timeout=15),
        )


class LoginAsManagementCommandTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create(username="user-1")

    @time_machine.travel(datetime(2023, 1, 1))
    def test_creates_token(self):
        token = ManagementAuthTokenGenerator().make_token(self.user.id)

        stdout = StringIO()
        call_command("login_as", self.user.username, stdout=stdout)

        self.assertIn(token, stdout.getvalue())

        self.client.get(stdout.getvalue())
        self.assertEqual(get_user(self.client), self.user)

    @override_settings(MANAGEMENT_AUTH_BASE_URL="https://management-auth.com")
    def test_get_absolute_url(self):
        stdout = StringIO()
        call_command("login_as", self.user.username, stdout=stdout)

        self.assertIn("https://management-auth.com", stdout.getvalue())

    @override_settings(WAGTAILADMIN_BASE_URL="https://wagtail.org")
    def test_get_absolute_url_from_wagtail(self):
        stdout = StringIO()
        call_command("login_as", self.user.username, stdout=stdout)

        self.assertIn("https://wagtail.org", stdout.getvalue())

    @override_settings(SITE_ID=1)
    def test_get_absolute_url_from_django_site(self):
        stdout = StringIO()
        call_command("login_as", self.user.username, stdout=stdout)

        self.assertIn("https://example.com", stdout.getvalue())



from django.core.signing import BadSignature, TimestampSigner
from django.utils.http import base36_to_int, int_to_base36


class ManagementAuthTokenGenerator:
    default_timeout = 60
    key_salt = __name__
    timeout_sep = "-"

    def __init__(self):
        self.signer = TimestampSigner(salt=self.key_salt)

    def make_token(self, user_id, timeout=None):
        if timeout is None:
            timeout = self.default_timeout

        return (
            int_to_base36(timeout)
            + self.timeout_sep
            + self.signer.sign_object([user_id, timeout], compress=True)
        )

    def get_user_from_token(self, token):
        if not token:
            return None

        # Parse the token
        try:
            ts_b36, token = token.split(self.timeout_sep, 1)
        except ValueError:
            return None

        try:
            max_age = base36_to_int(ts_b36)
        except ValueError:
            return None

        try:
            user_id, timeout = self.signer.unsign_object(token, max_age=max_age)
        except BadSignature:
            return None

        if max_age != timeout:
            return None

        return user_id



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.contenttypes",
    "tests",
    "management_auth",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
]

SECRET_KEY = "abcde12345"

ROOT_URLCONF = "tests.urls"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}



from django.conf import settings
from django.contrib.auth import get_user_model, login
from django.contrib.auth.views import redirect_to_login
from django.http.response import HttpResponseRedirect
from django.shortcuts import resolve_url
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.cache import never_cache

from .tokens import ManagementAuthTokenGenerator

User = get_user_model()


@method_decorator(never_cache, name="dispatch")
class LoginAsView(View):
    token_generator = ManagementAuthTokenGenerator()

    head = View.http_method_not_allowed

    def get_success_url(self):
        return resolve_url(settings.LOGIN_REDIRECT_URL)

    def get_error_response(self):
        return redirect_to_login("")

    def get(self, request, token):
        user_id = self.token_generator.get_user_from_token(token)

        if user_id is None:
            return self.get_error_response()

        user = User.objects.filter(id=user_id).first()

        if user is None:
            return self.get_error_response()

        user_last_login = user.last_login

        login(request, user)

        user.last_login = user_last_login
        user.save(update_fields=["last_login"])

        return HttpResponseRedirect(self.get_success_url())



from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.urls import reverse
from management_auth.tokens import ManagementAuthTokenGenerator

User = get_user_model()


class Command(BaseCommand):
    generator = ManagementAuthTokenGenerator()

    def add_arguments(self, parser):
        parser.add_argument("username", type=str)
        parser.add_argument(
            "--timeout", type=int, default=self.generator.default_timeout
        )

    def get_base_url(self):
        for setting in ["MANAGEMENT_AUTH_BASE_URL", "WAGTAILADMIN_BASE_URL"]:
            if base_url := getattr(settings, setting, None):
                return base_url

        if apps.is_installed("django.contrib.sites"):
            from django.contrib.sites.models import Site

            # Try Django's site framework
            try:
                return f"https://{Site.objects.get_current(None).domain}"
            except ImproperlyConfigured:
                pass

        return None

    def handle(self, *args, **options):
        user = User.objects.get(username=options["username"])

        token = self.generator.make_token(user.id, options["timeout"])

        url = reverse("management_auth_login_as", args=[token])

        if base_url := self.get_base_url():
            url = base_url + url

        self.stdout.writelines([url])
